// 2019-08-02 ffgl@subpixels.com
// - Replace create() calls with new operator

// 2019-10-23 ffgl@subpixels.com
// - Use ParamFloat
// - Introduce namespace ffgl::plugin::effect

#include "Gradient.h"
#include <ffglex/RGBAColor.h>
#include <ffglex/ScopedShaderBinding.h>
#include <ffglex/Utilities.h>
#include <math.h>//floor

using namespace ffglex;

namespace ffgl
{
namespace plugin
{
namespace effect
{

	static PluginInfo pluginInfo(
	PluginFactory< Gradient >,// Create method
#ifdef _DEBUG
	"?s01",             // Plugin unique ID of maximum length 4.
	"?Gradient Example",// Plugin name
#else
	"RS01",             // Plugin unique ID of maximum length 4.
	"Gradient Example", // Plugin name
#endif
	2,                             // API major version number
	1,                             // API minor version number
	1,                             // Plugin major version number
	000,                           // Plugin minor version number
	PluginType::FF_SOURCE,         // Plugin type
	"Sample FFGL Gradients plugin",// Plugin description
	"Resolume FFGL Example"        // About
);

// 2019-10-23 ffgl@subpixels.com
// - Use std::string instead of char[]
static const std::string vertexShaderCode = R"(#version 410 core
layout( location = 0 ) in vec4 vPosition;
layout( location = 1 ) in vec2 vUV;

out vec2 uv;

void main()
{
	gl_Position = vPosition;
	uv = vUV;
}
)";

// 2019-10-23 ffgl@subpixels.com
// - Renamed RGBLeft, RGBRight to rgb1, rgb2
// - Use std::string instead of char[]
static const std::string fragmentShaderCode = R"(#version 410 core
uniform vec3 rgb1;
uniform vec3 rgb2;

in vec2 uv;

out vec4 fragColor;

void main()
{
	fragColor = vec4( mix( rgb1, rgb2, uv.x ), 1.0 );
}
)";

// 2019-10-23 ffgl@subpixels.com - Use ParamFloat, ParameterType::FF_TYPE_HUE
Gradient::Gradient()
{
	// Input properties
	SetMinInputs( 0 );
	SetMaxInputs( 0 );

	// Parameters
	AddParam( hue1 = new ParamFloat( "Hue 1", ParameterType::FF_TYPE_HUE, "", 0.5f ) );
	AddParam( hue2 = new ParamFloat( "Hue 2", ParameterType::FF_TYPE_HUE, "", 0.0f ) );
	AddParam( saturation = new ParamFloat( "Saturation", "", 1.0f ) );
	AddParam( brightness = new ParamFloat( "Brightness", "", 1.0f ) );
}

// 2019-10-23 ffgl@subpixels.com
// - Use ScopedShaderBinding
// - Save uniform locations
FFResult Gradient::InitGL( const ViewportStruct* vp )
{
	if( !shader.Compile( vertexShaderCode, fragmentShaderCode ) )
	{
		DeinitGL();
		return FF_FAIL;
	}

	if( !quad.Initialise() )
	{
		DeinitGL();
		return FF_FAIL;
	}

	//FFGL requires us to leave the context in a default state on return, so use this scoped binding to help us do that.
	ScopedShaderBinding shaderBinding( shader.GetGLID() );
	rgb1Location = shader.FindUniform( "rgb1" );
	rgb2Location = shader.FindUniform( "rgb2" );

	// Use base-class init as success result so that it retains the viewport.
	return PluginBase::InitGL( vp );
}

// 2019-10-23 ffgl@subpixels.com
// - Avoid calling value getters multiple times
// - Use RGBAColor structs in place of int[3] arrays
// - Use ScopedShaderBinding
// - Use saved uniform locations
FFResult Gradient::ProcessOpenGL( ProcessOpenGLStruct* pGL )
{
	// We need to make sure the hue doesn't reach 1.0f, otherwise the result will be pink and not red how it should be
	float hue1Val = hue1->GetFloatValue();
	if( hue1Val == 1.0f )
		hue1Val = 0.0f;

	float hue2Val = hue2->GetFloatValue();
	if( hue2Val == 1.0f )
		hue2Val = 0.0f;

	float sat = saturation->GetFloatValue();
	float bri = brightness->GetFloatValue();

	// Translate from HSB to RGB
	RGBAColor rgb1;
	RGBAColor rgb2;

	HSVtoRGB( hue1Val, sat, bri, rgb1.r, rgb1.g, rgb1.b );
	HSVtoRGB( hue2Val, sat, bri, rgb2.r, rgb2.g, rgb2.b );

	// FFGL requires us to leave the context in a default state on return, so use this scoped binding to help us do that.
	ScopedShaderBinding shaderBinding( shader.GetGLID() );

	// Set shader uniforms
	shader.Set( rgb1Location, rgb1.r, rgb1.g, rgb1.b );
	shader.Set( rgb2Location, rgb2.r, rgb2.g, rgb2.b );

	// Render the quad using the shader
	quad.Draw();

	return FF_SUCCESS;
}

FFResult Gradient::DeinitGL()
{
	shader.FreeGLResources();
	quad.Release();

	return FF_SUCCESS;
}

}// namespace effect
}// namespace plugin
}// namespace ffgl
